import datetime

import parsel

from enext_exchange_py.mappers import (
    map_page_to_detailed_quote,
    map_page_to_factsheet,
    time_str_to_datetime,
)
from enext_exchange_py.models import DetailedQuote, Factsheet


def test__map_empty_page_to_detailed_quote():
    empty_page = parsel.Selector("")
    quote = map_page_to_detailed_quote(empty_page)

    assert isinstance(quote, DetailedQuote)
    assert quote.name is None
    assert quote.symbol is None
    assert quote.last_traded_price is None
    assert quote.last_traded_time is None
    assert quote.since_open_percentage is None
    assert quote.since_previous_close_percentage is None


def test__map_empty_page_to_factsheet():
    empty_page = parsel.Selector("")
    factsheet = map_page_to_factsheet(empty_page)

    assert isinstance(factsheet, Factsheet)
    assert factsheet.isin is None
    assert factsheet.code is None
    assert factsheet.instrument_type is None
    assert factsheet.instrument_sub_type is None
    assert factsheet.segment is None
    assert factsheet.trading_mode is None
    assert factsheet.trading_group is None
    assert factsheet.trading_currency is None
    assert factsheet.quantity_notation is None
    assert factsheet.shares_outstanding is None
    assert factsheet.tick_size is None


def test_time_str_to_datetime_tz_cet():
    time_str = "\t\t\t\t20/02/2026\t\t\t\t - 09:35 \n\t\t CET\n\t\t\t"
    date = time_str_to_datetime(time_str)
    assert date.year == 2026
    assert date.month == 2
    assert date.day == 20
    assert date.hour == 9
    assert date.minute == 35
    assert date.tzinfo is not None

    assert date.tzinfo.tzname(date) == "CET"
    assert date.tzinfo.utcoffset(date) == datetime.timedelta(hours=1)



from dataclasses import dataclass
from datetime import datetime

__all__ = ["DetailedQuote", "Quote", "EncryptedResponse", "Price", "Factsheet"]


@dataclass(frozen=True)
class EncryptedResponse:
    ct: str
    s: str
    iv: str


@dataclass(frozen=True)
class Price:
    value: float
    currency: str | None = None


@dataclass(frozen=True)
class DetailedQuote:
    name: str | None = None
    symbol: str | None = None
    last_traded_price: Price | None = None
    last_traded_time: datetime | None = None
    since_open_percentage: float | None = None
    since_previous_close_percentage: float | None = None


@dataclass(frozen=True)
class Quote:
    time: datetime
    price: float
    volume: int


@dataclass(frozen=True)
class Factsheet:
    isin: str | None = None
    code: str | None = None
    instrument_type: str | None = None
    instrument_sub_type: str | None = None
    segment: str | None = None
    trading_mode: str | None = None
    trading_group: str | None = None
    trading_currency: str | None = None
    quantity_notation: str | None = None
    shares_outstanding: int | None = None
    tick_size: str | None = None



import base64
import datetime
import hashlib
import json
from typing import Any, AsyncGenerator, Final, Generator

import httpx
from Crypto.Cipher import AES
from httpx import URL
from parsel import Selector

from enext_exchange_py.mappers import map_page_to_detailed_quote, map_page_to_factsheet
from enext_exchange_py.models import DetailedQuote, EncryptedResponse, Factsheet, Quote

__all__ = ["ExchangeClient"]

DEFAULT_BASE_URL: Final[str] = "https://live.euronext.com"
DEFAULT_HEADERS: Final[dict] = {
    "Cache-Control": "no-cache",
    "X-Requested-With": "XMLHttpRequest",
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:147.0) Gecko/20100101 Firefox/147.0",
}
DEFAULT_SECRET: Final[str] = "24ayqVo7yJma"


class ExchangeClient:
    _client: httpx.AsyncClient

    def __init__(
        self,
        base_url: URL | str = DEFAULT_BASE_URL,
        headers: dict | None = None,
        secret: str | bytes = DEFAULT_SECRET,
        language: str = "en",
        *args,
        **kwargs,
    ):
        if isinstance(base_url, str):
            base_url = URL(base_url)
        self.base_url = base_url

        if headers is None:
            self.headers = DEFAULT_HEADERS

        self._cookies = httpx.Cookies()
        self._client = httpx.AsyncClient(
            cookies=self._cookies,
            headers=headers,
            base_url=self.base_url,
            follow_redirects=True,
            *args,
            **kwargs,
        )
        if isinstance(secret, str):
            self._secret = secret.encode()
        self._language = language

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self._client.aclose()

    async def close(self):
        await self._client.aclose()

    @staticmethod
    def _obtain_key(password: bytes, salt: bytes, key_len: int, iv_len: int) -> tuple[bytes, bytes]:
        """
        Derives key and IV from password and salt. Uses PBKDF2 with md5 as a hash function.
        """
        dtot = b""
        d = b""
        while len(dtot) < (key_len + iv_len):
            # Hash current digest + password + salt
            d = hashlib.md5(d + password + salt).digest()
            dtot += d
        return dtot[:key_len], dtot[key_len : key_len + iv_len]

    @staticmethod
    def _decrypt_data(resp: EncryptedResponse, secret: bytes) -> Any:
        salt = bytes.fromhex(resp.s)
        ct = base64.b64decode(resp.ct)

        key, iv = ExchangeClient._obtain_key(secret, salt, 32, 16)

        cipher = AES.new(key, AES.MODE_CBC, iv)
        decrypted_bytes = cipher.decrypt(ct)
        # PKCS7 unpadding
        padding_len = decrypted_bytes[-1]
        if padding_len < 1 or padding_len > 16:
            # Fallback or error if padding is not valid PKCS7
            decrypted = decrypted_bytes.decode("utf-8")
        else:
            decrypted = decrypted_bytes[:-padding_len].decode("utf-8")

        return json.loads(decrypted)

    async def get_detailed_quote(self, symbol: str) -> DetailedQuote:
        path = f"/{self._language}/ajax/getDetailedQuote/{symbol}"

        resp = await self._client.get(path)
        page = Selector(resp.text)

        data = map_page_to_detailed_quote(page)
        return data

    @staticmethod
    def _parse_quotes_list(quotes_list: list[dict[str, Any]]) -> Generator[Quote, None, None]:
        for item in quotes_list:
            date_str = item.pop("time")
            item["time"] = datetime.datetime.strptime(date_str, "%Y-%m-%d %H:%M")
            yield Quote(**item)

    async def get_intraday_quotes(self, symbol: str) -> AsyncGenerator[Quote, None]:
        path = f"/intraday_chart/getChartData/{symbol}/intraday"

        resp = await self._client.get(path)
        enc_data = EncryptedResponse(**resp.json())
        data = self._decrypt_data(enc_data, self._secret)

        for value in self._parse_quotes_list(data):
            yield value

    async def get_historical_quotes(self, symbol: str) -> AsyncGenerator[Quote, None]:
        path = f"/intraday_chart/getChartData/{symbol}/max"

        resp = await self._client.get(path)
        enc_data = EncryptedResponse(**resp.json())
        data = self._decrypt_data(enc_data, self._secret)

        for value in self._parse_quotes_list(data):
            yield value

    async def get_factsheet(self, symbol: str) -> Factsheet:
        path = f"/{self._language}/ajax/getDetailedQuoteFactsheets/{symbol}"

        resp = await self._client.get(path)
        page = Selector(resp.text)

        data = map_page_to_factsheet(page)
        return data



import datetime
import unittest

import pytest

from enext_exchange_py.client import ExchangeClient
from enext_exchange_py.models import DetailedQuote, Quote


@pytest.mark.asyncio
class TestClient(unittest.IsolatedAsyncioTestCase):
    def __init__(self, *args, **kwargs):
        self.client = ExchangeClient()
        super().__init__(*args, **kwargs)

    async def test__get_detailed_quote(self):
        resp = await self.client.get_detailed_quote("LU0290358497-ETFP")

        assert isinstance(resp, DetailedQuote)
        assert resp.last_traded_price is not None
        assert resp.last_traded_price.value >= 0
        assert isinstance(resp.last_traded_time, datetime.datetime)
        assert resp.since_open_percentage is not None
        assert resp.since_previous_close_percentage is not None

    async def test__get_intraday_quotes(self):
        resp = self.client.get_intraday_quotes("LU0290358497-ETFP")
        first_value = await anext(resp)

        assert isinstance(first_value, Quote)
        assert isinstance(first_value.time, datetime.datetime)
        assert first_value.price > 0
        assert first_value.volume >= 0

    async def test__get_historical_quotes(self):
        resp = self.client.get_historical_quotes("LU0290358497-ETFP")
        first_value = await anext(resp)

        assert isinstance(first_value, Quote)
        assert isinstance(first_value.time, datetime.datetime)
        assert first_value.price > 0
        assert first_value.volume >= 0

    async def test__get_factsheet(self):
        resp = await self.client.get_factsheet("LU0290358497-ETFP")

        assert resp.isin == "LU0290358497"
        assert resp.code == "NSCIT0IXEON0"
        assert resp.instrument_type == "Trackers"
        assert resp.instrument_sub_type == "ETF"
        assert resp.trading_currency == "EUR"



import httpx

from enext_exchange_py.client import DEFAULT_SECRET, ExchangeClient
from enext_exchange_py.models import EncryptedResponse


def test__decrypt_data():
    url = "https://live.euronext.com/en/intraday_chart/getChartData/IE00BP3QZB59-ETFP/intraday"
    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36",
        "X-Requested-With": "XMLHttpRequest",
    }

    with httpx.Client(follow_redirects=True) as client:
        resp = client.get(url, headers=headers)
        resp.raise_for_status()
        data = resp.json()

    er = EncryptedResponse(ct=data["ct"], s=data["s"], iv=data["iv"])

    # Decrypt data
    decrypted = ExchangeClient._decrypt_data(er, DEFAULT_SECRET.encode())

    # Assertions
    assert isinstance(decrypted, list)
    assert len(decrypted) > 0
    # Check if first element has expected keys for intraday data
    first_point = decrypted[0]
    assert "time" in first_point
    assert "price" in first_point



import re
from datetime import datetime

from dateutil import tz
from parsel import Selector

from enext_exchange_py.models import DetailedQuote, Factsheet, Price

__all__ = ["time_str_to_datetime", "map_page_to_factsheet", "map_page_to_detailed_quote"]


# TODO: Replace this with a proper mapping, currently zoneinfo/datetime it is limited by the timezone data available on the system
TZ_MAPPING = {
    "UTC": tz.gettz("UTC"),
    "GMT": tz.gettz("Europe/London"),
    "BST": tz.gettz("Europe/London"),
    "CET": tz.gettz("Europe/Berlin"),
    "CEST": tz.gettz("Europe/Berlin"),
    "EET": tz.gettz("Europe/Helsinki"),
    "EEST": tz.gettz("Europe/Helsinki"),
}

CURRENCY_MAPPING = {
    "€": "EUR",
    "$": "USD",
}


def time_str_to_datetime(time_str: str) -> datetime:
    time_str = re.sub(r"[^A-Z\d/:]", " ", time_str)
    time_str = re.sub(r"\s+", " ", time_str)
    time_str = time_str.strip()

    parts = time_str.rsplit(" ", 1)
    dt_part = parts[0]
    tz_part = parts[1]

    date = datetime.strptime(dt_part, "%d/%m/%Y %H:%M")

    timezone = TZ_MAPPING.get(tz_part)

    date = date.replace(tzinfo=timezone)
    return date


def map_page_to_factsheet(page: Selector) -> Factsheet:
    table = page.css("table.table")

    raw_data = {}
    for row in table.css("tbody").css("tr"):
        cells = row.css("td::text").getall()
        if len(cells) != 2:
            continue
        key = cells[0].strip()
        value = cells[1].strip()

        if key is not None and value is not None:
            raw_data[key] = value.strip()

    data = {}
    if raw_data.get("ISIN") is not None:
        data["isin"] = raw_data["ISIN"].upper()

    if raw_data.get("Euronext Code") is not None:
        data["code"] = raw_data["Euronext Code"].upper()

    if raw_data.get("Intrument type") is not None:
        data["instrument_type"] = raw_data["Intrument type"]

    if raw_data.get("Intrument Sub-type") is not None:
        data["instrument_sub_type"] = raw_data["Intrument Sub-type"]

    if raw_data.get("Segment") is not None:
        data["segment"] = raw_data["Segment"]

    if raw_data.get("Trading Mode") is not None:
        data["trading_mode"] = raw_data["Trading Mode"]

    if raw_data.get("Trading Group") is not None:
        data["trading_group"] = raw_data["Trading Group"]

    if raw_data.get("Trading Currency") is not None:
        data["trading_currency"] = raw_data["Trading Currency"].upper()

    if raw_data.get("Quantity notation") is not None:
        data["quantity_notation"] = raw_data["Quantity notation"]

    if raw_data.get("Shared outstanding") is not None:
        data["shares_outstanding"] = raw_data["Shared outstanding"]

    if raw_data.get("Tick size") is not None:
        data["tick_size"] = raw_data["Tick size"]

    return Factsheet(**data)


def map_page_to_detailed_quote(page: Selector) -> DetailedQuote:
    name = page.css("h1").css("strong::text").get()
    if name is not None:
        name = name.strip()

    symbol = page.css("div.enx-symbol-top-custom::text").get()
    if symbol is not None:
        symbol = symbol.strip()

    last_traded_price_currency = page.css("#header-instrument-currency::text").get()
    if last_traded_price_currency is not None:
        last_traded_price_currency = last_traded_price_currency.strip()
        last_traded_price_currency = CURRENCY_MAPPING.get(
            last_traded_price_currency, last_traded_price_currency
        )

    last_traded_price = page.css("#header-instrument-price::text").get()
    if last_traded_price is not None:
        last_traded_price = float(last_traded_price.strip())
        last_traded_price = Price(last_traded_price, last_traded_price_currency)

    last_traded_time = page.css("div.last-price-date-time::text").get()
    if last_traded_time is not None:
        last_traded_time = time_str_to_datetime(last_traded_time)

    since_open_percentage = None
    since_previous_close_percentage = None
    data_header = page.css("div.data-header__row")
    for row in data_header.css("div.col"):
        text = row.css("div.data-12::text").get()
        if text is None:
            continue
        value = data_header.css("span.data-24::text").get()
        if value is None:
            continue
        value = re.sub(r"\s+", " ", value).strip()
        value = re.sub(r"[^0-9.]", "", value)
        if text == "Since Open":
            since_open_percentage = float(value)
        elif text == "Since Previous Close":
            since_previous_close_percentage = float(value)

    return DetailedQuote(
        name=name,
        symbol=symbol,
        last_traded_price=last_traded_price,
        last_traded_time=last_traded_time,
        since_open_percentage=since_open_percentage,
        since_previous_close_percentage=since_previous_close_percentage,
    )
